@@ -1,13 +1,15 @@
 use anyhow::{Context, Result};
+use deqp_runner::deqp_command::DeqpCommand;
 use deqp_runner::mock_deqp::{mock_deqp, MockDeqp};
-use deqp_runner::JunitGeneratorOptions;
 use deqp_runner::{
-    parallel_deqp, parse_regex_set, read_lines, Deqp, DeqpCommand, RunnerResults, TestCase,
+    parallel_test, parse_regex_set, process_results, read_baseline, read_lines,
+    CommandLineRunOptions, RunnerResults, TestCase, TestCommand,
 };
+use deqp_runner::{JunitGeneratorOptions, TestConfiguration};
 use log::*;
 use serde::Deserialize;
 use std::collections::HashMap;
-use std::fs::{self, File};
+use std::fs::File;
 use std::path::{Path, PathBuf};
 use std::time::Duration;
 use structopt::StructOpt;
@@ -40,19 +42,6 @@ enum SubCommand {
     MockDeqp(MockDeqp),
 }
 
-fn parse_key_val<T, U>(s: &str) -> Result<(T, U), Box<dyn std::error::Error>>
-where
-    T: std::str::FromStr,
-    T::Err: std::error::Error + 'static,
-    U: std::str::FromStr,
-    U::Err: std::error::Error + 'static,
-{
-    let pos = s
-        .find('=')
-        .ok_or_else(|| format!("invalid KEY=value: no `=` found in `{}`", s))?;
-    Ok((s[..pos].parse()?, s[pos + 1..].parse()?))
-}
-
 #[derive(Debug, StructOpt)]
 pub struct Run {
     #[structopt(long, help = "path to deqp binary")]
@@ -61,65 +50,8 @@ pub struct Run {
     #[structopt(long, help = "path to deqp caselist (such as *-mustpass.txt)")]
     caselist: Vec<PathBuf>,
 
-    #[structopt(long = "output", help = "path to output directory")]
-    output_dir: PathBuf,
-
-    #[structopt(
-        long,
-        help = "path to baseline results (such as output/failures.csv from another run)"
-    )]
-    baseline: Option<PathBuf>,
-
-    #[structopt(
-        long,
-        help = "path to file of regexes of tests to skip running (for runtime or stability reasons)"
-    )]
-    skips: Option<Vec<PathBuf>>,
-
-    #[structopt(
-        short = "t",
-        long = "include-tests",
-        help = "regexes of tests to include (non-matching tests are skipped)"
-    )]
-    include: Vec<String>,
-
-    #[structopt(
-        long,
-        help = "path to file of regexes of tests to assume any failures in those tests are flaky results (but still run them, for long-term status tracking)"
-    )]
-    flakes: Option<Vec<PathBuf>>,
-
-    #[structopt(
-        long,
-        default_value = "60.0",
-        help = "per-test timeout in floating point seconds"
-    )]
-    timeout: f32,
-
-    #[structopt(
-        short = "j",
-        long,
-        default_value = "0",
-        help = "Number of dEQP processes to invoke in parallel (default 0 = number of CPUs in system)"
-    )]
-    jobs: usize,
-
-    #[structopt(
-        long,
-        default_value = "25",
-        help = "Number of fails or flakes to print in the summary line (0 = no limit)"
-    )]
-    summary_limit: usize,
-
-    #[structopt(long, default_value = "1", help = "Runs 1 out of every N tests.")]
-    fraction: usize,
-
-    #[structopt(
-        long,
-        default_value = "1",
-        help = "Skips the first N-1 tests in the test list before applying --fraction (useful for running N/M fraciton of the test list across multiple devices)."
-    )]
-    fraction_start: usize,
+    #[structopt(flatten)]
+    common: CommandLineRunOptions,
 
     #[structopt(
         long,
@@ -128,11 +60,30 @@ pub struct Run {
     )]
     tests_per_group: usize,
 
+    // The runner  corehas clever code I inherited from the parallel-deqp-runner
+    // project to, near the end of the list of test of tests to shard out, make
+    // smaller groups of tests so that you don't end up with a long test list
+    // left in one deqp process limiting your total runtime.
+    //
+    // The problem that cleverness faces is this table of deqp startup time:
+    //
+    //             freedreno  radeon
+    // deqp-gles2  0.2s       0.08s
+    // deqp-vk     2.0s       0.6s
+    //
+    // Even if all the slowest tests I have (a few deqp-vks on radeon at 6-7s)
+    // land in the same test list near the end of the run, you're still going to
+    // end up spawning a ton of extra deqp processes trying to be clever, costing
+    // more than your tests.
+    //
+    // So, disable it by default by using our normal group size the whole time
+    // unless someone sets the option to something else.
     #[structopt(
         long,
-        help = "Minimum number of tests to scale down to in each deqp invocation (defaults to tests_per_group)"
+        default_value = "0",
+        help = "Minimum number of tests to scale down to in each deqp invocation (defaults to 0 to match tests_per_group)"
     )]
-    min_tests_per_group: Option<usize>,
+    min_tests_per_group: usize,
 
     #[structopt(
         long,
@@ -140,13 +91,6 @@ pub struct Run {
     )]
     testlog_to_xml: Option<PathBuf>,
 
-    #[structopt(
-        parse(try_from_str = parse_key_val),
-        long = "env",
-        help = "Environment variables to set when invoking the dEQP process"
-    )]
-    env: Vec<(String, String)>,
-
     #[structopt(
         long,
         default_value = "",
@@ -182,70 +126,14 @@ pub struct Suite {
     #[structopt(long, help = "Path to suite.toml")]
     suite: PathBuf,
 
-    #[structopt(
-        long,
-        help = "path to baseline results (such as output/failures.csv from another run)"
-    )]
-    baseline: Option<PathBuf>,
-
-    #[structopt(
-        long,
-        help = "path to file of regexes of tests to skip running (for runtime or stability reasons)"
-    )]
-    skips: Option<Vec<PathBuf>>,
-
-    #[structopt(
-        long,
-        help = "path to file of regexes of tests to assume any failures in those tests are flaky results (but still run them, for long-term status tracking)"
-    )]
-    flakes: Option<Vec<PathBuf>>,
-
-    #[structopt(
-        short = "t",
-        long = "include-tests",
-        help = "regexes of tests to include (non-matching tests are skipped)"
-    )]
-    include: Vec<String>,
-
-    #[structopt(long, default_value = "1", help = "Runs 1 out of every N tests.")]
-    fraction: usize,
-
-    #[structopt(
-        long,
-        default_value = "1",
-        help = "Skips the first N-1 tests in the test list before applying --fraction (useful for running N/M fraciton of the test list across multiple devices)."
-    )]
-    fraction_start: usize,
-
-    #[structopt(long = "output", help = "path to output directory")]
-    output_dir: PathBuf,
-
-    #[structopt(
-        short = "j",
-        long,
-        default_value = "0",
-        help = "Number of dEQP processes to invoke in parallel (default 0 = number of CPUs in system)"
-    )]
-    jobs: usize,
-
-    #[structopt(
-        long,
-        default_value = "25",
-        help = "Number of fails or flakes to print in the summary line (0 = no limit)"
-    )]
-    summary_limit: usize,
+    #[structopt(flatten)]
+    common: CommandLineRunOptions,
 
     #[structopt(
         long,
         help = "Optional path to executor/testlog-to-xml, for converting QPA files to usable XML"
     )]
     testlog_to_xml: Option<PathBuf>,
-
-    #[structopt(
-        parse(try_from_str = parse_key_val),
-        help = "Environment variables to set when invoking the dEQP process"
-    )]
-    env: Vec<(String, String)>,
 }
 
 // Common structure for configuring a deqp run between Run (single run) and Suite (muliple Runs)
@@ -274,7 +162,7 @@ struct DeqpConfig {
     timeout: f32,
 
     #[serde(default)]
-    min_tests_per_group: Option<usize>,
+    min_tests_per_group: usize,
 
     #[serde(default)]
     tests_per_group: usize,
@@ -300,36 +188,18 @@ struct DeqpConfig {
 
 impl DeqpConfig {
     fn deqp(&self, output_dir: &Path, qpa_to_xml: Option<PathBuf>) -> Result<DeqpCommand> {
-        let baseline = match &self.baseline {
-            Some(path) => {
-                let mut file = File::open(path).context("Reading baseline")?;
-                RunnerResults::from_csv(&mut file)?
-            }
-            None => RunnerResults::new(),
-        };
-
-        let skips = if !self.skips.is_empty() {
-            Some(deqp_runner::parse_regex_set(read_lines(&self.skips)?)?)
-        } else {
-            None
-        };
-
-        let flakes = if !self.flakes.is_empty() {
-            Some(deqp_runner::parse_regex_set(read_lines(&self.flakes)?)?)
-        } else {
-            None
-        };
-
         Ok(DeqpCommand {
             deqp: self.deqp.clone(),
             args: self.deqp_args.clone(),
-            output_dir: output_dir.to_path_buf(),
-            skips,
-            flakes,
-            baseline,
-            timeout: Duration::from_secs_f32(self.timeout),
+            config: TestConfiguration {
+                output_dir: output_dir.to_path_buf(),
+                skips: parse_regex_set(read_lines(&self.skips)?)?,
+                flakes: parse_regex_set(read_lines(&self.flakes)?)?,
+                baseline: read_baseline(self.baseline.as_ref())?,
+                timeout: Duration::from_secs_f32(self.timeout),
+                env: self.env.clone(),
+            },
             qpa_to_xml,
-            env: self.env.clone(),
             prefix: self.prefix.to_owned(),
         })
     }
@@ -363,42 +233,6 @@ impl DeqpConfig {
             .filter(|test| include_filters.iter().all(|x| x.is_match(test.name())))
             .collect::<Vec<TestCase>>();
 
-        let tests_per_group = usize::max(
-            1,
-            usize::min(
-                (test_names.len() + rayon_threads - 1) / rayon_threads,
-                self.tests_per_group,
-            ),
-        );
-
-        let min_tests_per_group = if let Some(min_tests_per_group) = self.min_tests_per_group {
-            if min_tests_per_group == 0 {
-                eprintln!("--min_tests_per_group must be >= 1.");
-                std::process::exit(1);
-            }
-            min_tests_per_group
-        } else {
-            // This runner has clever code I inherited from the parallel-deqp-runner
-            // project to, near the end of the list of test of tests to shard out, make
-            // smaller groups of tests so that you don't end up with a long test list
-            // left in one deqp process limiting your total runtime.
-            //
-            // The problem that cleverness faces is this table of deqp startup time:
-            //
-            //             freedreno  radeon
-            // deqp-gles2  0.2s       0.08s
-            // deqp-vk     2.0s       0.6s
-            //
-            // Even if all the slowest tests I have (a few deqp-vks on radeon at 6-7s)
-            // land in the same test list near the end of the run, you're still going to
-            // end up spawning a ton of extra deqp processes trying to be clever, costing
-            // more than your tests.
-            //
-            // So, disable it by default by using our normal group size the whole time
-            // unless someone sets the option to something else.
-            self.tests_per_group
-        };
-
         if !test_names.is_empty()
             && (!self.renderer_check.is_empty() || !self.version_check.is_empty())
         {
@@ -463,12 +297,20 @@ impl DeqpConfig {
             }
         }
 
+        let groups =
+            deqp.split_tests_to_groups(test_names, self.tests_per_group, self.min_tests_per_group)?;
+
         println!(
             "Running dEQP on {} threads in {}-test groups",
-            rayon_threads, tests_per_group
+            rayon_threads,
+            if let Some((_, tests)) = groups.get(0) {
+                tests.len()
+            } else {
+                0
+            }
         );
 
-        Ok(deqp.split_tests_to_groups(test_names, tests_per_group, min_tests_per_group))
+        Ok(groups)
     }
 }
 
@@ -477,32 +319,6 @@ struct SuiteConfig {
     deqp: Vec<DeqpConfig>,
 }
 
-fn process_results(results: &RunnerResults, output_dir: &Path, summary_limit: usize) -> Result<()> {
-    results.write_results(&mut File::create(&output_dir.join("results.csv"))?)?;
-    results.write_failures(&mut File::create(&output_dir.join("failures.csv"))?)?;
-
-    results.print_summary(if summary_limit == 0 {
-        std::usize::MAX
-    } else {
-        summary_limit
-    });
-
-    if !results.is_success() {
-        std::process::exit(1);
-    }
-
-    Ok(())
-}
-
-fn set_jobs(jobs: usize) {
-    if jobs > 0 {
-        rayon::ThreadPoolBuilder::new()
-            .num_threads(jobs)
-            .build_global()
-            .unwrap();
-    }
-}
-
 fn main() -> Result<()> {
     stderrlog::new().module(module_path!()).init().unwrap();
 
@@ -510,59 +326,39 @@ fn main() -> Result<()> {
 
     match opts.subcmd {
         SubCommand::Run(run) => {
-            set_jobs(run.jobs);
-
-            fs::create_dir_all(&run.output_dir).context("creating output directory")?;
-
-            if run.fraction < 1 {
-                eprintln!("fraction must be >= 1.");
-                std::process::exit(1);
-            }
-            if run.fraction_start < 1 {
-                eprintln!("fraction_start must be >= 1.");
-                std::process::exit(1);
-            }
+            run.common.setup()?;
 
             let config = DeqpConfig {
                 deqp: run.deqp,
                 deqp_args: run.deqp_args,
                 caselists: run.caselist,
-                baseline: run.baseline,
-                include: run.include,
-                timeout: run.timeout,
-                skips: run.skips.unwrap_or_else(Vec::new),
-                flakes: run.flakes.unwrap_or_else(Vec::new),
+                baseline: run.common.baseline,
+                include: run.common.include,
+                timeout: run.common.timeout,
+                skips: run.common.skips,
+                flakes: run.common.flakes,
                 tests_per_group: run.tests_per_group,
                 min_tests_per_group: run.min_tests_per_group,
-                fraction_start: run.fraction_start,
-                fraction: run.fraction,
-                env: run.env.into_iter().collect(),
+                fraction_start: run.common.fraction_start,
+                fraction: run.common.fraction,
+                env: run.common.env.into_iter().collect(),
                 prefix: "".to_owned(),
                 renderer_check: run.renderer_check,
                 version_check: run.version_check,
             };
-            let deqp = config.deqp(&run.output_dir, run.testlog_to_xml)?;
+            let deqp = config.deqp(&run.common.output_dir, run.testlog_to_xml)?;
 
-            let results = parallel_deqp(std::io::stdout(), config.test_groups(&deqp, &[])?)?;
-            process_results(&results, &run.output_dir, run.summary_limit)?;
+            let results = parallel_test(std::io::stdout(), config.test_groups(&deqp, &[])?)?;
+            process_results(&results, &run.common.output_dir, run.common.summary_limit)?;
         }
 
         SubCommand::Suite(suite) => {
-            set_jobs(suite.jobs);
+            suite.common.setup()?;
 
             let toml_str = std::fs::read_to_string(&suite.suite).context("Reading config TOML")?;
             let suite_config =
                 toml::from_str::<SuiteConfig>(toml_str.as_str()).context("Parsing config TOML")?;
 
-            if suite.fraction < 1 {
-                eprintln!("fraction must be >= 1.");
-                std::process::exit(1);
-            }
-            if suite.fraction_start < 1 {
-                eprintln!("fraction_start must be >= 1.");
-                std::process::exit(1);
-            }
-
             // Apply defaults to the configs.
             let mut configs = Vec::new();
             for mut config in suite_config.deqp {
@@ -576,19 +372,15 @@ fn main() -> Result<()> {
                     config.fraction_start = 1;
                 }
 
-                if let Some(files) = suite.skips.as_ref() {
-                    for f in files {
-                        config.skips.push(f.clone());
-                    }
+                for f in &suite.common.skips {
+                    config.skips.push(f.clone());
                 }
 
-                if let Some(files) = suite.flakes.as_ref() {
-                    for f in files {
-                        config.flakes.push(f.clone());
-                    }
+                for f in &suite.common.flakes {
+                    config.flakes.push(f.clone());
                 }
 
-                if let Some(run_baseline) = suite.baseline.as_ref() {
+                if let Some(run_baseline) = suite.common.baseline.as_ref() {
                     if config.baseline.is_some() {
                         eprintln!(
                             "baseline may only be set on either the command line or per-deqp."
@@ -599,14 +391,14 @@ fn main() -> Result<()> {
                 }
 
                 // Apply global fraction to the suite's internal fraction.
-                config.fraction *= suite.fraction;
-                config.fraction_start += suite.fraction_start - 1;
+                config.fraction *= suite.common.fraction;
+                config.fraction_start += suite.common.fraction_start - 1;
 
                 if config.timeout == 0.0 {
                     config.timeout = 60.0;
                 }
 
-                for (var, data) in &suite.env {
+                for (var, data) in &suite.common.env {
                     config.env.insert(var.to_owned(), data.to_owned());
                 }
 
@@ -615,17 +407,21 @@ fn main() -> Result<()> {
 
             let mut deqp = Vec::new();
             for config in &configs {
-                deqp.push(config.deqp(&suite.output_dir, suite.testlog_to_xml.clone())?);
+                deqp.push(config.deqp(&suite.common.output_dir, suite.testlog_to_xml.clone())?);
             }
 
             let mut test_groups = Vec::new();
             for (config, deqp) in configs.iter().zip(deqp.iter()) {
-                let mut groups = config.test_groups(deqp, &suite.include)?;
+                let mut groups = config.test_groups(deqp, &suite.common.include)?;
                 test_groups.append(&mut groups);
             }
 
-            let results = parallel_deqp(std::io::stdout(), test_groups)?;
-            process_results(&results, &suite.output_dir, suite.summary_limit)?;
+            let results = parallel_test(std::io::stdout(), test_groups)?;
+            process_results(
+                &results,
+                &suite.common.output_dir,
+                suite.common.summary_limit,
+            )?;
         }
 
         SubCommand::Junit(junit) => {
