@@ -285,6 +285,11 @@ fn parse_piglit_command(
     }))
 }
 
+/* Replace ',' to avoid conflicts in the csv file */
+pub fn piglit_sanitize_test_name(test: &str) -> String {
+    test.replace(",", "-")
+}
+
 pub fn parse_piglit_xml_testlist(
     folder: &Path,
     file_content: &str,
@@ -307,11 +312,8 @@ pub fn parse_piglit_xml_testlist(
     }
 
     for test in doc.descendants().filter(|n| n.has_tag_name("Test")) {
-        /* Replace ',' to avoid conflicts in the csv file */
-        let test_name = test
-            .attribute("name")
-            .context("getting test name")?
-            .replace(",", "-");
+        let test_name =
+            piglit_sanitize_test_name(test.attribute("name").context("getting test name")?);
 
         match test.attribute("type") {
             // multi_shader implements the same feature as deqp-runner: grouping test runs in a
