@@ -656,6 +656,29 @@ dEQP-GLES2.info.version,Pass"
             "1:02:15"
         );
     }
+
+    fn add_result(results: &mut RunnerResults, test: &str, status: RunnerStatus) {
+        results.record_result(RunnerResult {
+            test: test.to_string(),
+            status,
+            duration: Duration::new(0, 0),
+            subtest: false,
+        });
+    }
+
+    #[test]
+    fn results_is_success() {
+        let mut results = RunnerResults::new();
+
+        add_result(&mut results, "pass1", RunnerStatus::Pass);
+        add_result(&mut results, "pass2", RunnerStatus::Pass);
+
+        assert_eq!(results.is_success(), true);
+
+        add_result(&mut results, "Crash", RunnerStatus::Crash);
+
+        assert_eq!(results.is_success(), false);
+    }
 }
 
 #[derive(Debug, StructOpt)]
