@@ -2,17 +2,44 @@ use super::*;
 use ::deqp_runner::{RunnerResults, RunnerStatus};
 use anyhow::{Context, Result};
 use std::io::Cursor;
+use std::path::{Path, PathBuf};
 use std::process::{Command, Stdio};
 
+enum MaybeTempFile {
+    Temp(tempfile::TempPath),
+    File(PathBuf),
+}
+
+impl AsRef<Path> for MaybeTempFile {
+    fn as_ref(&self) -> &Path {
+        match self {
+            MaybeTempFile::Temp(t) => t,
+            MaybeTempFile::File(path) => path,
+        }
+    }
+}
+
+pub fn add_includes(cmd: &mut Command, includes: &[String]) {
+    if !includes.is_empty() {
+        cmd.arg("--include-tests");
+        for i in includes {
+            cmd.arg(i);
+        }
+    }
+}
+
 /// Builder for a mocked deqp-runner invocation
 #[derive(Default)]
 struct DeqpMock {
-    pub caselists: Vec<tempfile::TempPath>,
+    pub caselists: Vec<MaybeTempFile>,
     pub baselines: Vec<tempfile::TempPath>,
     pub skips: Vec<tempfile::TempPath>,
     pub flakes: Vec<tempfile::TempPath>,
     pub runner_args: Vec<String>,
     pub prefix: String,
+    pub renderer_check: String,
+    pub version_check: String,
+    pub includes: Vec<String>,
 }
 
 impl DeqpMock {
@@ -39,11 +66,11 @@ impl DeqpMock {
         let child = child.arg(output_dir.path());
 
         let child = child.arg("--timeout");
-        let child = child.arg("1");
+        let mut child = child.arg("1");
 
         for caselist_file in &self.caselists {
             child.arg("--caselist");
-            child.arg(&caselist_file);
+            child.arg(caselist_file.as_ref());
         }
 
         for baseline_file in &self.baselines {
@@ -61,6 +88,18 @@ impl DeqpMock {
             child.arg(&flakes_file);
         }
 
+        if !self.renderer_check.is_empty() {
+            child.arg("--renderer-check");
+            child.arg(&self.renderer_check);
+        }
+
+        if !self.version_check.is_empty() {
+            child.arg("--version-check");
+            child.arg(&self.version_check);
+        }
+
+        add_includes(&mut child, &self.includes);
+
         for arg in &self.runner_args {
             child.arg(arg);
         }
@@ -94,8 +133,14 @@ impl DeqpMock {
         S: AsRef<str>,
         I: IntoIterator<Item = S>,
     {
-        self.caselists
-            .push(lines_tempfile(lines).context("writing caselist").unwrap());
+        self.caselists.push(MaybeTempFile::Temp(
+            lines_tempfile(lines).context("writing caselist").unwrap(),
+        ));
+        self
+    }
+
+    pub fn with_caselist_file(&mut self, file: &Path) -> &mut DeqpMock {
+        self.caselists.push(MaybeTempFile::File(file.to_path_buf()));
         self
     }
 
@@ -135,6 +180,20 @@ impl DeqpMock {
         self.prefix = arg.to_owned();
         self
     }
+
+    pub fn with_includes(&mut self, arg: &str) -> &mut DeqpMock {
+        self.includes.push(arg.to_owned());
+        self
+    }
+
+    pub fn with_renderer_check(&mut self, arg: &str) -> &mut DeqpMock {
+        self.renderer_check = arg.to_owned();
+        self
+    }
+    pub fn with_version_check(&mut self, arg: &str) -> &mut DeqpMock {
+        self.version_check = arg.to_owned();
+        self
+    }
 }
 
 fn mocked_deqp_runner<S: AsRef<str>>(tests: Vec<S>) -> RunnerResults {
@@ -153,17 +212,19 @@ fn result_status<S: AsRef<str>>(results: &RunnerResults, test: S) -> RunnerStatu
 #[derive(Default)]
 struct DeqpSuite {
     deqps: Vec<DeqpMock>,
+    includes: Vec<String>,
 }
 
-fn write_file_list_toml<W: Write>(
+fn write_file_list_toml<W: Write, P: AsRef<Path>>(
     file: &mut W,
     key: &str,
-    files: &[tempfile::TempPath],
+    files: &[P],
 ) -> Result<()> {
     if !files.is_empty() {
         write!(file, "{} = [", key).context("writing file list")?;
         for f in files {
-            write!(file, "\"{}\"", f.display()).context("writing file list")?;
+            let path = f.as_ref().to_str().context("formatting path")?;
+            write!(file, "\"{}\"", path).context("writing file list")?;
         }
         writeln!(file, "]").context("writing file list")?;
     }
@@ -198,6 +259,20 @@ impl DeqpSuite {
                     writeln!(toml, r#"baseline = "{}""#, deqp.baselines[0].display())?;
                 }
 
+                if !deqp.renderer_check.is_empty() {
+                    writeln!(toml, r#"renderer_check = "{}""#, &deqp.renderer_check)?;
+                }
+                if !deqp.version_check.is_empty() {
+                    writeln!(toml, r#"version_check = "{}""#, &deqp.version_check)?;
+                }
+
+                if !deqp.includes.is_empty() {
+                    write!(toml, r#"include = ["#,)?;
+                    for i in &deqp.includes {
+                        write!(toml, r#""{}", "#, i)?;
+                    }
+                    writeln!(toml, "]")?;
+                }
                 // Passed as the first arg of the "deqp" binary (the deqp-runner we passed as "deqp ="!) to trigger its mock-deqp mode
                 writeln!(toml, r#"deqp_args = ["mock-deqp"]"#)?;
 
@@ -227,7 +302,9 @@ impl DeqpSuite {
         let child = child.arg(output_dir.path());
 
         let child = child.arg("--suite");
-        let child = child.arg(toml.path());
+        let mut child = child.arg(toml.path());
+
+        add_includes(&mut child, &self.includes);
 
         let output = child
             .spawn()
@@ -254,6 +331,11 @@ impl DeqpSuite {
         self.deqps.push(deqp);
         self
     }
+
+    pub fn with_includes(&mut self, arg: &str) -> &mut DeqpSuite {
+        self.includes.push(arg.to_owned());
+        self
+    }
 }
 
 #[test]
@@ -541,8 +623,7 @@ fn missing_flakes() {
 #[test]
 fn includes() {
     let results = DeqpMock::new()
-        .with_runner_arg("-t")
-        .with_runner_arg("dEQP-GLES2.test.p.*")
+        .with_includes("dEQP-GLES2.test.p.*")
         .with_cases(vec!["dEQP-GLES2.test.p.1", "dEQP-GLES2.test.f.2"])
         .run()
         .unwrap()
@@ -553,11 +634,34 @@ fn includes() {
     assert_eq!(results.result_counts.total, 1);
 }
 
+#[test]
+fn caselist_subdir() {
+    let mut caselist = PathBuf::from(env!("CARGO_MANIFEST_DIR"));
+    caselist.push("resources/test/caselist-subdir.txt");
+
+    let results = DeqpMock::new().with_caselist_file(&caselist).run().unwrap();
+
+    println!("stdout: {}", results.stdout);
+    println!("stderr: {}", results.stderr);
+    let results = results.results.unwrap();
+
+    assert_eq!(results.result_counts.pass, 2);
+    assert_eq!(results.result_counts.fail, 0);
+    assert_eq!(results.result_counts.total, 2);
+    assert_eq!(
+        result_status(&results, "dEQP-GLES2.p.1"),
+        RunnerStatus::Pass
+    );
+    assert_eq!(
+        result_status(&results, "dEQP-GLES2.p.2"),
+        RunnerStatus::Pass
+    );
+}
+
 #[test]
 fn bad_includes() {
     let results = DeqpMock::new()
-        .with_runner_arg("-t")
-        .with_runner_arg("*")
+        .with_includes("*")
         .with_cases(vec!["dEQP-GLES2.test.p"])
         .run()
         .unwrap();
@@ -698,3 +802,170 @@ fn suite_prefix_skips() {
         RunnerStatus::Skip
     );
 }
+
+#[test]
+fn suite_missing() {
+    let mut deqp = DeqpMock::new();
+    deqp.with_cases(vec!["dEQP-GLES2.test.m.1"]);
+    deqp.with_prefix("variant-");
+
+    let results = DeqpSuite::new().with_deqp(deqp).run().unwrap();
+    assert_eq!(Some(1), results.status.code());
+    assert!(results.stdout.contains("Missing: 1"));
+
+    let results = results.results.unwrap();
+    assert_eq!(results.result_counts.missing, 1);
+    assert_eq!(results.result_counts.fail, 0);
+
+    assert_eq!(
+        results.tests["variant-dEQP-GLES2.test.m.1"].status,
+        RunnerStatus::Missing
+    );
+}
+
+#[test]
+fn suite_includes() {
+    let mut deqp = DeqpMock::new();
+    deqp.with_cases(vec![
+        "dEQP-GLES2.test.p.1.a.b",
+        "dEQP-GLES2.test.p.1.a.c",
+        "dEQP-GLES2.test.p.1.a.d",
+        "dEQP-GLES2.test.p.1.a.e",
+    ]);
+    deqp.with_includes(".a");
+
+    let results = DeqpSuite::new()
+        .with_deqp(deqp)
+        .with_includes(".c")
+        .run()
+        .unwrap();
+    assert_eq!(Some(0), results.status.code());
+    let results = results.results.unwrap();
+    assert_eq!(results.result_counts.pass, 1);
+    assert_eq!(
+        results.tests["dEQP-GLES2.test.p.1.a.c"].status,
+        RunnerStatus::Pass
+    );
+}
+
+#[test]
+fn renderer_version_check() {
+    let mut deqp = DeqpMock::new();
+    deqp.with_cases(vec!["dEQP-GLES2.test.p.1"]);
+    deqp.with_renderer_check("Intel.*CFL");
+    deqp.with_version_check("OpenGL ES 3.2.*git-6a274846ba");
+
+    let results = deqp.run().unwrap();
+
+    assert!(results
+        .stdout
+        .contains("renderer: Mesa Intel(R) UHD Graphics 630 (CFL GT2)"));
+    assert!(results
+        .stdout
+        .contains("version: OpenGL ES 3.2 Mesa 21.3.0-devel (git-6a274846ba)"));
+
+    assert_eq!(Some(0), results.status.code());
+
+    let results = results.results.unwrap();
+    assert_eq!(results.result_counts.pass, 1);
+    assert_eq!(results.result_counts.fail, 0);
+}
+
+#[test]
+fn renderer_check_fail() {
+    let mut deqp = DeqpMock::new();
+    deqp.with_cases(vec!["dEQP-GLES2.test.p.1"]);
+    deqp.with_renderer_check("AMD");
+    deqp.with_version_check("OpenGL ES 3.2.*git-6a274846ba");
+
+    let results = deqp.run().unwrap();
+
+    assert!(results
+        .stdout
+        .contains("renderer: Mesa Intel(R) UHD Graphics 630 (CFL GT2)"));
+    assert!(results
+        .stdout
+        .contains("version: OpenGL ES 3.2 Mesa 21.3.0-devel (git-6a274846ba)"));
+    assert_eq!(Some(1), results.status.code());
+}
+
+#[test]
+fn version_check_fail() {
+    let mut deqp = DeqpMock::new();
+    deqp.with_cases(vec!["dEQP-GLES2.test.p.1"]);
+    deqp.with_renderer_check("Intel.*CFL");
+    deqp.with_version_check("OpenGL ES 3.1.*git-6a274846ba");
+
+    let results = deqp.run().unwrap();
+
+    assert!(results
+        .stdout
+        .contains("renderer: Mesa Intel(R) UHD Graphics 630 (CFL GT2)"));
+    assert!(results
+        .stdout
+        .contains("version: OpenGL ES 3.2 Mesa 21.3.0-devel (git-6a274846ba)"));
+    assert_eq!(Some(1), results.status.code());
+}
+
+#[test]
+fn vk_renderer_version_check() {
+    let mut deqp = DeqpMock::new();
+    deqp.with_cases(vec!["dEQP-VK.test.p.1"]);
+    deqp.with_renderer_check("AMD RADV");
+
+    let results = deqp.run().unwrap();
+
+    assert!(results.stdout.contains("renderer: AMD RADV VEGA10"));
+
+    assert_eq!(Some(0), results.status.code());
+
+    let results = results.results.unwrap();
+    assert_eq!(results.result_counts.pass, 1);
+    assert_eq!(results.result_counts.fail, 0);
+}
+
+#[test]
+fn vk_renderer_check_fail() {
+    let mut deqp = DeqpMock::new();
+    deqp.with_cases(vec!["dEQP-VK.test.p.1"]);
+    deqp.with_renderer_check("Intel");
+
+    let results = deqp.run().unwrap();
+
+    assert!(results.stdout.contains("renderer: AMD RADV VEGA10"));
+    assert_eq!(Some(1), results.status.code());
+}
+
+#[test]
+fn suite_renderer_version_check() {
+    let mut deqp = DeqpMock::new();
+    deqp.with_cases(vec!["dEQP-GLES2.test.p.1"]);
+    deqp.with_renderer_check("Intel.*CFL");
+    deqp.with_version_check("OpenGL ES 3.2.*git-6a274846ba");
+
+    let results = DeqpSuite::new().with_deqp(deqp).run().unwrap();
+
+    assert!(results
+        .stdout
+        .contains("renderer: Mesa Intel(R) UHD Graphics 630 (CFL GT2)"));
+    assert_eq!(Some(0), results.status.code());
+
+    let results = results.results.unwrap();
+    assert_eq!(results.result_counts.pass, 1);
+    assert_eq!(results.result_counts.fail, 0);
+}
+
+#[test]
+fn suite_renderer_check_fail() {
+    let mut deqp = DeqpMock::new();
+    deqp.with_cases(vec!["dEQP-GLES2.test.p.1"]);
+    deqp.with_renderer_check("AMD");
+    deqp.with_version_check("OpenGL ES 3.2.*git-6a274846ba");
+
+    let results = DeqpSuite::new().with_deqp(deqp).run().unwrap();
+
+    assert!(results
+        .stdout
+        .contains("renderer: Mesa Intel(R) UHD Graphics 630 (CFL GT2)"));
+    assert_eq!(Some(1), results.status.code());
+}
